/**
 * Copyright (C) 2013-2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

/**
 * Long term plan:
 * Eventually, this file and its header will no longer be a part of this project.
 * The features made available by this core will slowly be transferred away in pieces
 * to other locations, such as the internal Jupiter library, or various plugins.
 *
 * There is a relatively thin line between what will go in Jupiter, and what will be
 * pushed to plugins.
 * Example: The "Command" and "Plugin" classes ended up in Jupiter; they're abstract concepts
 * that doesn't put a specific behavior on any mechanisms. The "IRCCommand" class, however,
 * alters the behavior of the IRC Client, and thus should be contained within a plugin.
 */

#include <cstdio>
#include <cstring>
#include "IRC_Bot.h"
#include "INIFile.h"
#include "Plugin.h"
#include "Functions.h"
#include "IRC_Command.h"

IRC_Bot::IRC_Bot(const char *configSection) : Client(configSection)
{
	IRC_Bot::commandPrefix = this->readConfigValue("Prefix");
	if (IRC_Bot::commandPrefix == nullptr) IRC_Bot::commandPrefix = "";
	IRC_Bot::commandPrefixLen = strlen(IRC_Bot::commandPrefix);
	for (int i = IRCMasterCommandList->size() - 1; i >= 0; i--) IRC_Bot::addCommand(IRCMasterCommandList->get(i)->copy());
	IRC_Bot::setCommandAccessLevels();
}

IRC_Bot::~IRC_Bot()
{
	while (IRC_Bot::commands.size()) delete IRC_Bot::commands.remove(0);
}

void IRC_Bot::addCommand(IRCCommand *cmd)
{
	IRC_Bot::commands.add(cmd);
}

bool IRC_Bot::freeCommand(const char *trigger)
{
	for (int i = IRC_Bot::commands.size() - 1; i >= 0; i--)
	{
		if (IRC_Bot::commands.get(i)->matches(trigger))
		{
			delete IRC_Bot::commands.remove(i);
			return true;
		}
	}
	return false;
}

int IRC_Bot::getCommandIndex(const char *trigger)
{
	for (int i = IRC_Bot::commands.size() - 1; i >= 0; i--) if (IRC_Bot::commands.get(i)->matches(trigger)) return i;
	return -1;
}

IRCCommand *IRC_Bot::getCommand(const char *trigger)
{
	short i = IRC_Bot::getCommandIndex(trigger);
	if (i < 0) return nullptr;
	return IRC_Bot::commands.get(i);
}

Jupiter::ArrayList<IRCCommand> IRC_Bot::getAccessCommands(Channel *chan, char access)
{
	Jupiter::ArrayList<IRCCommand> r;
	for (int i = IRC_Bot::commands.size() - 1; i >= 0; i--)
	{
		IRCCommand *cmd = IRC_Bot::commands.get(i);
		if (cmd->getAccessLevel(chan) == access) r.add(cmd);
	}
	return r;
}

Jupiter::CStringL IRC_Bot::getTriggers(Jupiter::ArrayList<IRCCommand> cmds)
{
	Jupiter::CStringL r;
	for (unsigned int i = 0; i < cmds.size(); i++)
	{
		r += cmds[i]->getTrigger();
		r += ' ';
	}
	return r;
}

#include <cctype> // isdigit

void IRC_Bot::setCommandAccessLevels()
{
	// Rewrite this later
	// Note: Prepare for abstraction of configuration file type.
	char *commandSection = new char[strlen(this->getConfigSection()) + 16];
	strcpy(commandSection, "DefaultCommands");
	int sIndex = Config->getSectionIndex(commandSection);
	if (sIndex >= 0)
	{
		int sLen = Config->getSectionLength(sIndex);
		for (int i = 0; i < sLen; i++)
		{
			const char *key = Config->getKey(commandSection, i);
			int pos = findSymbol(key, '.', 0);
			if (pos >= 0)
			{
				char *command = charToChar(key, 0, pos);
				if (command != nullptr)
				{
					IRCCommand *cmd = IRC_Bot::getCommand(command);
					if (cmd != nullptr)
					{
						if (isdigit(key[pos+1])) cmd->setAccessLevel(atoi(key+pos+1), Config->getInt(commandSection, key));
						else cmd->setAccessLevel(key+pos+1, Config->getInt(commandSection, key));
					}
					else if (this->getPrintOutput()) printf("Unable to find command \"%s\"" ENDL, key);
					free(command);
				}
			}
			else
			{
				IRCCommand *cmd = IRC_Bot::getCommand(key);
				if (cmd != nullptr) cmd->setAccessLevel(Config->getInt(commandSection, key));
				else if (this->getPrintOutput()) printf("Unable to find command \"%s\"" ENDL, key);
			}
		}
	}
	sprintf(commandSection, "%sCommands", this->getConfigSection());
	sIndex = Config->getSectionIndex(commandSection);
	if (sIndex >= 0)
	{
		int sLen = Config->getSectionLength(sIndex);
		for (int i = 0; i < sLen; i++)
		{
			const char *key = Config->getKey(commandSection, i);
			int pos = findSymbol(key, '.', 0);
			if (pos >= 0)
			{
				char *command = charToChar(key, 0, pos);
				if (command != nullptr)
				{
					IRCCommand *cmd = IRC_Bot::getCommand(command);
					if (cmd != nullptr)
					{
						if (isdigit(key[pos+1])) cmd->setAccessLevel(atoi(key+pos+1), Config->getInt(commandSection, key));
						else cmd->setAccessLevel(key+pos+1, Config->getInt(commandSection, key));
					}
					else if (this->getPrintOutput()) printf("Unable to find command \"%s\"" ENDL, key);
					free(command);
				}
			}
			else
			{
				IRCCommand *cmd = IRC_Bot::getCommand(key);
				if (cmd != nullptr) cmd->setAccessLevel(Config->getInt(commandSection, key));
				else if (this->getPrintOutput()) printf("Unable to find command \"%s\"" ENDL, key);
			}
		}
	}
	delete[] commandSection;
}

int IRC_Bot::rehash(const char *filename)
{
	int r;
	if (filename != nullptr)
	{
		Config->flushData();
		r = Config->readFile(filename);
	}
	else r = Config->reload();
	IRC_Bot::setCommandAccessLevels();
	for (int i = Jupiter::plugins->size() - 1; i >= 0; i--) Jupiter::plugins->get(i)->OnRehash(this);
	return r;
}

void IRC_Bot::OnChat(const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringType &message)
{
	short chanIndex = this->getChannelIndex(channel.c_str());
	if (chanIndex >= 0)
	{
		Channel *chan = this->getChannel(chanIndex);
		if (chan->getType() >= 0)
		{
			bool matchesPrefix = true;
			size_t i;
			for (i = 0; i < IRC_Bot::commandPrefixLen; i++)
			{
				if (toupper(message.c_str()[i]) != toupper(IRC_Bot::commandPrefix[i]))
				{
					matchesPrefix = false;
					break;
				}
			}
			i += strspn(message.c_str() + i, WHITESPACE);
			if (matchesPrefix)
			{
				// Make this more efficient later. (Add static GetWord to CString_Strict and CString_Loose)
				Jupiter::CStringS smessage = message;
				Jupiter::CStringS cword = smessage.getWord(0, WHITESPACE);
				const char *command = cword.c_str()+1;
				Jupiter::CStringS parameters = smessage.gotoWord(1, WHITESPACE); //message.substring(i + cword.size(), message.size());
				IRCCommand *cmd = IRC_Bot::getCommand(command);
				if (cmd != nullptr)
				{
					int cAccess = cmd->getAccessLevel(chan);
					if (cAccess >= 0 && Jupiter::IRC::Client::getAccessLevel(channel.c_str(), nick.c_str()) >= cAccess) cmd->trigger(this, channel, nick, parameters);
					else this->sendNotice(nick.c_str(), "Access Denied.");
				}
			}
		}
	}
}

int IRC_Bot::think()
{
	int r = Jupiter::IRC::Client::think();
	for (int i = Jupiter::plugins->size() - 1; i >= 0; i--) Jupiter::plugins->get(i)->OnThink(this);
	return r;
}