/**
 * Copyright (C) 2013-2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#if !defined _IRC_BOT_H_HEADER
#define _IRC_BOT_H_HEADER

/**
 * @file IRC_Bot.h
 * @brief Provides extendable bot-like interfacing with the IRC client.
 */

#include "Jupiter_Bot.h"
#include "IRC_Client.h"
#include "ArrayList.h"

/** Forward Declaration */
class IRCCommand;

/** Template class */
template class JUPITER_BOT_API Jupiter::ArrayList<IRCCommand>;

/**
* @brief Provies extendable bot-like interfacing with the IRC client.
*/
class JUPITER_BOT_API IRC_Bot : public Jupiter::IRC::Client
{
public:

	/**
	* @brief Pulls the configured access levels from the config and applies them.
	*/
	void setCommandAccessLevels();

	/**
	* @brief Adds a command to the command list.
	*/
	void addCommand(IRCCommand *cmd);

	/**
	* @brief Removes a command from the command list.
	*
	* @param trigger Trigger of the command to remove
	* @return True if a command is removed, false otherwise.
	*/
	bool freeCommand(const char *trigger);

	/**
	* @brief Gets the index of a command.
	*
	* @param trigger Trigger of the command to find.
	* @return Index of the first match for the trigger if found, -1 otherwise.
	*/
	int getCommandIndex(const char *trigger);

	/**
	* @return Gets a command.
	*
	* @param trigger Trigger of the command to find.
	* @return First command using the specified trigger, nullptr otherwise.
	*/
	IRCCommand *getCommand(const char *trigger);

	/**
	* @brief Creates and returns an ArrayList of IRC Commands with a specified access level.
	*
	* @param chan Channel which access levels are set.
	* @param access Access level to match.
	* @return An ArrayList containing pointers to all of the matching commands.
	*/
	Jupiter::ArrayList<IRCCommand> getAccessCommands(Jupiter::IRC::Client::Channel *chan, char access);

	/**
	* @brief Gets the triggers of all the commands in an ArrayList, and adds them to a space-deliminated string.
	*
	* @param cmds Commands to construct the string with.
	* @return A string containing the triggers of the commands in a space-deliminated list.
	*/
	static Jupiter::CStringL getTriggers(Jupiter::ArrayList<IRCCommand> cmds);

	/**
	* @brief Rehashes the connection settings.
	*
	* @param filename Name of file from which to get configuration options. Reloads current config if nullptr.
	* @return Number of configuration options set within the config on success; -1 otherwise.
	*/
	int rehash(const char *filename = nullptr);

	/**
	* @brief Calls Jupiter::IRC::Client::think() and triggers OnThink event for plugins.
	*
	* @return Result of Jupiter::IRC::Client::think().
	*/
	int think();

	/** Constructor for IRC_Bot */
	IRC_Bot(const char *configSection);

	/** Destructor for IRC_Bot */
	~IRC_Bot();

	/** Overloaded events from Jupiter::IRC::Client */
protected:

	void OnChat(const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringType &message);

	/** Private members for internal usage */
private:
	Jupiter::ArrayList<IRCCommand> commands;
	const char *commandPrefix;
	size_t commandPrefixLen;
};

#endif // _IRC_BOT_H_HEADER