/**
 * Copyright (C) 2013-2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <exception>
#include <thread>
#include "Functions.h"
#include "INIFile.h"
#include "Queue.h"
#include "Socket.h"
#include "Plugin.h"
#include "IRC_Bot.h"
#include "ServerManager.h"
#include "Console_Command.h"
#include "IRC_Command.h"

int nPlugins = 0;
Jupiter::Queue inputQueue;

void onTerminate()
{
	puts("Terminate signal received...");
}

void onExit()
{
	puts("Exit signal received; Cleaning up...");
	Jupiter::Socket::cleanup();
	puts("Clean-up complete. Closing...");
}

void inputLoop()
{
	const size_t inputSize = 1024;
	char *input = new char[inputSize];
	while (1)
	{
		fgets(input, inputSize, stdin);
		input[strcspn(input, "\r\n")] = 0;
		inputQueue.enqueue(input);
	}
	delete[] input;
}

int main(int argc, const char **args)
{
	atexit(onExit);
	std::set_terminate(onTerminate);
	// I'm having an issue where threading crashes on Linux...
#if defined _WIN32
	std::thread inputThread(inputLoop);
#endif // _WIN32
	srand((unsigned int) time(0));
	puts(Jupiter::copyright);
	const char *configFileName = CONFIG_INI;
	for (int i = 1; i < argc; i++)
	{
		if (streqli(args[i], "-help"))
		{
			puts("Help coming soon, to a theatre near you!");
			return 0;
		}
		else if (streqli(args[i], "-config") && ++i < argc) configFileName = args[i];
		else if (streqli(args[i], "-pluginsdir") && ++i < argc) Jupiter::setPluginDirectory(args[i]);
		else if (streqli(args[i], "-configFormat") && ++i < argc) puts("Feature not yet supported!");
		else printf("Warning: Unknown command line argument \"%s\" specified. Ignoring...", args[i]);
	}
	puts("Loading config file...");
	if (!Jupiter::IRC::Client::Config->readFile(configFileName))
	{
		puts("Unable to read config file. Closing...");
		exit(0);
	}

	fputs("Config loaded. ", stdout);
	const char *pDir = Jupiter::IRC::Client::Config->get("Config", "PluginsDirectory");
	if (pDir != nullptr)
	{
		Jupiter::setPluginDirectory(pDir);
		printf("Plugins will be loaded from \"%s\"." ENDL, pDir);
	}

	puts("Loading plugins...");
	const char *pluginList = Jupiter::IRC::Client::Config->get("Config", "Plugins");
	if (pluginList == nullptr) puts("No plugins to load!");
	else
	{
		nPlugins = wordCount(pluginList);
		printf("Attempting to load %d plugins..." ENDL, nPlugins);
		for (short i = 0; i < nPlugins; i++)
		{
			char *tPlug = getWord(pluginList, i);
			if (Jupiter::loadPlugin(tPlug) == nullptr) fprintf(stderr, "WARNING: Failed to load plugin \"%s\"!" ENDL, tPlug);
			else printf("\"%s\" loaded successfully." ENDL, tPlug);
			free(tPlug);
		}
	}

	if (consoleCommands->size() > 0) printf("%u Console Commands have been initialized%s" ENDL, consoleCommands->size(), getConsoleCommand("help") == nullptr ? "." : "; type \"help\" for more information.");
	if (IRCMasterCommandList->size() > 0) printf("%u IRC Commands have been loaded into the master list." ENDL, IRCMasterCommandList->size());

	puts("Retreiving network list...");
	const char *serverList = Jupiter::IRC::Client::Config->get("Config", "Servers");
	if (serverList == nullptr)
	{
		puts("Unable to find network list.");
		exit(0);
	}
	int nServers = wordCount(serverList);
	printf("Attempting to connect to %d servers..." ENDL, nServers);
	for (short i = 0; i < nServers; i++)
	{
		char *tDir = getWord(serverList, i);
		serverManager->addServer(tDir);
		free(tDir);
	}
	puts("Sockets established.");
	while (1)
	{
		for (int i = Jupiter::plugins->size() - 1; i >= 0; i--) if (Jupiter::plugins->get(i)->think() != 0) Jupiter::freePlugin(i);
		serverManager->think();
		const char *input = (const char *)inputQueue.dequeue();
		if (input != nullptr)
		{
			char *command = getWord(input, 0);
			if (command != nullptr)
			{
				ConsoleCommand *cmd = getConsoleCommand(command);
				size_t commandLength = strlen(command);
				if (cmd != nullptr) cmd->trigger(input + commandLength + strspn(input + commandLength, " \t\r\n"));
				else printf("Error: Command \"%s\" not found." ENDL, command);
			}
		}
		std::this_thread::sleep_for((std::chrono::milliseconds(1)));
	}
	return 0;
}
