/**
 * Copyright (C) 2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#include <cstring> // strlen
#include "CoreCommands.h"
#include "IRC_Bot.h"
#include "INIFile.h"
#include "Functions.h"
#include "ArrayList.h"

// Help Console Command

HelpConsoleCommand::HelpConsoleCommand()
{
	this->addTrigger("help");
}

void HelpConsoleCommand::trigger(const char *parameters)
{
	if (streql(parameters, ""))
	{
		fputs("Supported commands:", stdout);
		for (int i = consoleCommands->size() - 1; i >= 0; i--)
		{
			fputc(' ', stdout);
			fputs(consoleCommands->get(i)->getTrigger(), stdout);
		}
		printf(ENDL "%s - %s" ENDL, Jupiter::version, Jupiter::copyright);
		puts("For command-specific help, use: help <command>");
	}
	else
	{
		const char *t = getWord(parameters, 0);
		ConsoleCommand *cmd = getConsoleCommand(t);
		if (cmd) puts(cmd->getHelp());
		else printf("Error: Command \"%s\" not found." ENDL, t);
		delete[] t;
	}
}

const char *HelpConsoleCommand::getHelp()
{
	return "Lists commands, or sends command-specific help. Syntax: help [command]";
}

CONSOLE_COMMAND_INIT(HelpConsoleCommand)

// Version Console Command

VersionConsoleCommand::VersionConsoleCommand()
{
	this->addTrigger("version");
	this->addTrigger("versioninfo");
	this->addTrigger("copyright");
	this->addTrigger("copyrightinfo");
	this->addTrigger("client");
	this->addTrigger("clientinfo");
	this->addTrigger("info");
}

void VersionConsoleCommand::trigger(const char *parameters)
{
	fprintf(stdout, "Version: %s" ENDL "%s" ENDL, Jupiter::version, Jupiter::copyright);
}

const char *VersionConsoleCommand::getHelp()
{
	return "Displays version and copyright information";
}

CONSOLE_COMMAND_INIT(VersionConsoleCommand)

// Help Command.

void HelpIRCCommand::create()
{
	this->addTrigger("help");
}

void HelpIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	short cIndex = source->getChannelIndex(channel.c_str());
	if (cIndex >= 0)
	{
		Jupiter::IRC::Client::Channel *chan = source->getChannel(source->getChannelIndex(channel.c_str()));
		char access = source->getAccessLevel(channel.c_str(), nick.c_str());
		if (parameters == nullptr)
		{
			for (char i = 0; i <= access; i++)
			{
				Jupiter::ArrayList<IRCCommand> cmds = source->getAccessCommands(chan, i);
				if (cmds.size() > 0)
				{
					Jupiter::CStringL triggers = source->getTriggers(cmds);
					if (triggers.size() >= 0)
					{
						char *r = new char[triggers.size() + 64];
						sprintf(r, "Access level %d Commands: %s", i, triggers.c_str());
						source->sendNotice(nick.c_str(), r);
						delete[] r;
					}
				}
			}
			source->sendNotice(nick.c_str(), "For command-specific help, use: help <command>");
		}
		else
		{
			const char *t = getWord(parameters.c_str(), 0);
			IRCCommand *cmd = source->getCommand(t);
			if (cmd)
			{
				if (access >= cmd->getAccessLevel(chan)) source->sendNotice(nick.c_str(), cmd->getHelp());
				else source->sendNotice(nick.c_str(), "Access Denied.");
			}
			else source->sendNotice(nick.c_str(), "Error: Command not found.");
			delete[] t;
		}
	}
}

const char *HelpIRCCommand::getHelp()
{
	return "Syntax: help [command]";
}

IRC_COMMAND_INIT(HelpIRCCommand)

// Sync Command

void SyncIRCCommand::create()
{
	this->addTrigger("sync");
	this->setAccessLevel(4);
}

void SyncIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (source->Config == nullptr) source->sendMessage(channel.c_str(), "Unable to find Config data.");
	else
	{
		bool r;
		if (parameters.size() > 0) r = source->Config->sync(parameters.c_str());
		else r = source->Config->sync();
		if (r) source->sendMessage(channel.c_str(), "Config data synced to file successfully.");
		else source->sendMessage(channel.c_str(), "Unable to sync Config data.");
	}
}

const char *SyncIRCCommand::getHelp()
{
	return "Syntax: sync [Filename]";
}

IRC_COMMAND_INIT(SyncIRCCommand)

// Rehash Command

void RehashIRCCommand::create()
{
	this->addTrigger("rehash");
	this->setAccessLevel(4);
}

void RehashIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (source->Config == nullptr) source->sendMessage(channel.c_str(), "Unable to find Config data.");
	else
	{
		int r = source->rehash(parameters.c_str());
		if (r >= 0) source->sendMessage(channel.c_str(), "Config data has been reloaded from the file successfully.");
		else source->sendMessage(channel.c_str(), "There was an error trying to reload the Config data. This is potentially fatal.");
	}
}

const char *RehashIRCCommand::getHelp()
{
	return "Syntax: rehash [Filename]";
}

IRC_COMMAND_INIT(RehashIRCCommand)

// Plugin instantiation and entry point.
CoreCommandsPlugin pluginInstance;

extern "C" __declspec(dllexport) Jupiter::Plugin *getPlugin()
{
	return &pluginInstance;
}
