/**
 * Copyright (C) 2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#include "ExtraCommands.h"
#include "IRC_Bot.h"
#include "Functions.h"
#include <cstring>

// Raw Console Command

RawConsoleCommand::RawConsoleCommand()
{
	this->addTrigger("raw");
	this->addTrigger("sendraw");
}

void RawConsoleCommand::trigger(const char *parameters)
{
	if (wordCount(parameters) < 2)
	{
		puts("Error: Too few parameters.");
		return;
	}
	const char *network = getWord(parameters, 0);
	size_t networkLen = strlen(network);
	const char *message = parameters + networkLen + strspn(parameters + networkLen, WHITESPACE);
	unsigned int total = 0;
	for (unsigned int i = 0; i < serverManager->size(); i++)
	{
		IRC_Bot *server = serverManager->getServer(i);
		if (strmatchi(network, server->getConfigSection()))
		{
			server->send(message);
			total++;
		}
	}
	delete[] network;
	printf("%u packets sent." ENDL, total);
}

const char *RawConsoleCommand::getHelp()
{
	return "Syntax: raw <network> <message>";
}

CONSOLE_COMMAND_INIT(RawConsoleCommand)

// Message Console Command

MessageConsoleCommand::MessageConsoleCommand()
{
	this->addTrigger("ircmsg");
	this->addTrigger("message");
	this->addTrigger("privmsg");
}

void MessageConsoleCommand::trigger(const char *parameters)
{
	if (wordCount(parameters) < 3)
	{
		puts("Error: Too few parameters.");
		return;
	}
	const char *network = getWord(parameters, 0);
	const char *dest = getWord(parameters, 1);
	size_t networkLen = strlen(network);
	size_t destLen = strlen(dest);
	size_t tmpSz = strspn(parameters + networkLen, WHITESPACE);
	const char *message = parameters + networkLen + tmpSz + destLen + strspn(parameters + networkLen + tmpSz + destLen, WHITESPACE);
	unsigned int total = 0;
	for (unsigned int i = 0; i < serverManager->size(); i++)
	{
		IRC_Bot *server = serverManager->getServer(i);
		if (strmatchi(network, server->getConfigSection()))
		{
			server->sendMessage(dest, message);
			total++;
		}
	}
	delete[] dest;
	delete[] network;
	printf("%u messages sent." ENDL, total);
}

const char *MessageConsoleCommand::getHelp()
{
	return "Sends a message to an IRC user/channel, on a specified network (accepts wildcards). Syntax: msg <network> <destination> <message>";
}

CONSOLE_COMMAND_INIT(MessageConsoleCommand)

// Join Command

void JoinIRCCommand::create()
{
	this->addTrigger("Join");
	this->setAccessLevel(3);
}

void JoinIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (parameters.size() > 0)
	{
		if (wordCount(parameters.c_str()) == 1) source->joinChannel(parameters.c_str());
		else source->joinChannel(parameters.getWord(0, WHITESPACE).c_str(), parameters.getWord(1, WHITESPACE).c_str());
	}
	else source->sendNotice(nick.c_str(), "Error: Too Few Parameters. Syntax: join <channel> [password]");
}

const char *JoinIRCCommand::getHelp()
{
	return "Joins a channel. Syntax: join <Channel>";
}

IRC_COMMAND_INIT(JoinIRCCommand)

// Part Command

void PartIRCCommand::create()
{
	this->addTrigger("Part");
	this->setAccessLevel(3);
}

void PartIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (parameters.size() > 0)
	{
		if (parameters.wordCount(WHITESPACE) == 1) source->partChannel(parameters.c_str());
		else source->partChannel(parameters.getWord(0, WHITESPACE).c_str(), parameters.c_str() + findSymbol(parameters.c_str(), ' ', 0) + 1);
	}
	else source->partChannel(channel.c_str());
}

const char *PartIRCCommand::getHelp()
{
	return "Parts from a channel. Syntax: part [channel=here] [message=empty]";
}

IRC_COMMAND_INIT(PartIRCCommand)

// Info Command

void InfoIRCCommand::create()
{
	this->addTrigger("info");
	this->setAccessLevel(4);
}

void InfoIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	char response[256];
	short channelCount = source->getChannelsSize();
	sprintf(response, "Outputing data for %d channels...", source->getChannelCount());
	source->sendMessage(channel.c_str(), response);
	for (short a = 0; a < channelCount; a++)
	{
		Jupiter::IRC::Client::Channel *chan = source->getChannel(a);
		if (chan != nullptr)
		{
			sprintf(response, "Channel %s - Type: %d", chan->getName(), chan->getType());
			source->sendMessage(channel.c_str(), response);
			for (short b = 0; b < chan->getUserCount(); b++)
			{
				sprintf(response, "PRIVMSG %s :User %s (prefix: %c) of channel %s" ENDL, channel.c_str(), chan->getUser(b), chan->getUserPrefix(b) ? chan->getUserPrefix(b) : ' ', chan->getName());
				source->send(response);
			}
		}
	}
}

const char *InfoIRCCommand::getHelp()
{
	return "TEMPORARY COMMAND - Spits out some info about channels. Syntax: info";
}

IRC_COMMAND_INIT(InfoIRCCommand)

// Exit command

void ExitIRCCommand::create()
{
	this->addTrigger("exit");
	this->setAccessLevel(5);
}

void ExitIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	source->sendMessage(channel.c_str(), "Bye!");
	exit(0);
}

const char *ExitIRCCommand::getHelp()
{
	return "Syntax: exit";
}

IRC_COMMAND_INIT(ExitIRCCommand)

// IRC Connect command

void IRCConnectIRCCommand::create()
{
	this->addTrigger("IRCConnect");
	this->addTrigger("IRCReconnect");
	this->setAccessLevel(5);
}

void IRCConnectIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (parameters == nullptr)
	{
		char *cSection = makestr(source->getConfigSection());
		source->disconnect("Connect command used; reconnecting...");
		source->disconnect(true);
		serverManager->addServer(cSection);
		delete[] cSection;
	}
	else
	{
		IRC_Bot *server = serverManager->getServer(parameters.c_str());
		if (server != nullptr)
		{
			source->sendMessage(channel.c_str(), "Warning: Server already exists. Severing connection...");
			source->disconnect(true);
		}
		if (serverManager->addServer(parameters.c_str())) source->sendMessage(channel.c_str(), "Connection successfully established; server added to server list.");
		else source->sendMessage(channel.c_str(), "Error: Unable to find configuration settings for server, or connection refused.");
	}
}

const char *IRCConnectIRCCommand::getHelp()
{
	return "Connects/reconnects to an IRC server, based on config entry. Syntax: IRCConnect [server=here]";
}

IRC_COMMAND_INIT(IRCConnectIRCCommand)

// IRC Disconnect command

void IRCDisconnectIRCCommand::create()
{
	this->addTrigger("IRCDisconnect");
	this->setAccessLevel(5);
}

void IRCDisconnectIRCCommand::trigger(IRC_Bot *source, const Jupiter::CStringType &channel, const Jupiter::CStringType &nick, const Jupiter::CStringS &parameters)
{
	if (parameters == nullptr) source->disconnect("Disconnect command used", true);
	else
	{
		IRC_Bot *server = serverManager->getServer(parameters.c_str());
		if (server == nullptr) source->sendNotice(nick.c_str(), "Error: Server not found.");
		else server->disconnect("Disconnect command used remotely", true);
	}
}

const char *IRCDisconnectIRCCommand::getHelp()
{
	return "Disconnects from an IRC server, based on config entry. Syntax: IRCDisconnect [server=here]";
}

IRC_COMMAND_INIT(IRCDisconnectIRCCommand)

// Plugin instantiation and entry point.
FunCommandsPlugin pluginInstance;

extern "C" __declspec(dllexport) Jupiter::Plugin *getPlugin()
{
	return &pluginInstance;
}
