/**
 * Copyright (C) 2013-2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#if !defined _INIFILE_H_HEADER
#define _INIFILE_H_HEADER

/**
 * @file INIFile.h
 * @brief Provides interaction with INI files.
 */

#include "Jupiter.h"

namespace Jupiter
{
	/**
	* @brief Provides interfacing with INI files.
	* INIFile objects store the contents of .INI files into memory,
	* and access them as neccessary. The INIFile object can store
	* the contents of multiple files at a time, but does not store
	* where data came from. As such, you can read from multiple files,
	* however the contents will always go to a single file when synced.
	* Disk contents are only modified when sync() is called.
	*/
	class JUPITER_API INIFile
	{
	private:
		struct Data;
		Data *data_;
	public:
		INIFile();
		~INIFile();

		/**
		* @brief Flushes all stored data.
		*/
		void flushData();

		/**
		* @brief Reads data from a file.
		* The readFile function does not currently support keys which
		* are not under a section, though this is not currently an issue,
		* and may still be added later. The parser will also automatically
		* remove any excessive whitespace, including any whitespace prefixing
		* a key name, prefixing a key value, trailing a key name, or traling
		* a key value.
		*
		* @param fileName String containing the name of a file.
		* @return The number of entries added on success, -1 otherwise.
		*/
		int readFile(const char *fileName);

		/**
		* @brief Flushes all stored data and recollects from the last read file.
		*
		* @return The number of entries added on success, -1 otherwise.
		*/
		int reload();

		/**
		* @brief Syncs data to a file.
		*
		* @param fileName String containing the name of a file.
		* @return True on success, false otherwise.
		*/
		bool sync(const char *fileName);

		/**
		* @brief Syncs data to the file that was last read from.
		*
		* @return True on success, false otherwise.
		*/
		bool sync();

		/**
		* @brief Sets a key's value. Adds a key if an existing one is not found. Does not modify any files.
		*
		* @param section Section within which the data is to be stored.
		* @param keyName String containing key name.
		* @param keyValue String containing new key value.
		* @return True if a new key was added, false otherwise.
		*/
		bool set(const char *section, const char *keyName, const char *keyValue);

		/**
		* @brief Removes a key. Does not modify any files.
		*
		* @param section String containing section.
		* @param keyName String containing key name.
		* @return Number of entries removed.
		*/
		int remove(const char *section, const char *keyName);

		/**
		* @brief Returns the number of sections in memory.
		*
		* @return Number of sections.
		*/
		int getSections();

		/**
		* @brief Returns a section.
		*
		* @param index Index of the section to get.
		* @return String containing section if it exists, nullptr otherwise.
		*/
		const char *getSection(int index);

		/**
		* @brief Returns the index of a section.
		*
		* @param section The name of the section.
		* @return The index of section if it exists, -1 otherwise.
		*/
		int getSectionIndex(const char *section);

		/**
		* @brief Returns the number of keys in a section.
		*
		* @return Number of data entries if it exists, -1 otherwise.
		*/
		int getSectionLength(int index);

		/**
		* @brief Returns a key name.
		*
		* @param section String containing section.
		* @param keyIndex Index of the key.
		* @return Key name on success, nullptr otherwise.
		*/
		const char *getKey(const char *section, int keyIndex);

		/**
		* @brief Returns the index of a key.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return Index of key if it exists, -1 otherwise.
		*/
		int getKeyIndex(const char *section, const char *key);

		/**
		* @brief Returns the value of a key.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return Pointer to a string containing key value if it exists, nullptr otherwise.
		*/
		const char *get(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a boolean.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return true if key exists and is not "false" or "0", false otherwise.
		*/
		bool getBool(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a short.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return short value of the key if it exits, 0 otherwise.
		*/
		short getShort(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as an int.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return int value of the key if it exits, 0 otherwise.
		*/
		int getInt(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a long.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return long value of the key if it exits, 0 otherwise.
		*/
		long getLong(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a long long.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return long long value of the key if it exits, 0 otherwise.
		*/
		//long long getLongLong(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a float.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return float value of the key if it exits, 0 otherwise.
		*/
		float getFloat(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a double.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return double value of the key if it exits, 0 otherwise.
		*/
		double getDouble(const char *section, const char *key);

		/**
		* @brief Translates and returns the value of a key as a long double.
		*
		* @param section String containing section.
		* @param key String containing key name.
		* @return long double value of the key if it exits, 0 otherwise.
		*/
		//long double getLongDouble(const char *section, const char *key);
	};

}

#endif // _INIFILE_H_HEADER