/**
 * Copyright (C) 2013-2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#if !defined _IRC_CLIENT_H_HEADER
#define _IRC_CLIENT_H_HEADER

/**
 * @file IRC Client.h
 * @brief Provides connectivity to IRC servers.
 */

#include <cstdlib>
#include <cstdio>
#include "Jupiter.h"
#include "Thinker.h"
#include "IRC.h"
#include "CString.h"

#define CONFIG_INI "Config.ini" /** Default location of the Config file. */

namespace Jupiter
{

	class INIFile; /** Forward declaration for Jupiter::INIFile. See Jupiter::INIFile in INIFile.h. */

	namespace IRC
	{

		/**
		* @brief Provides connectivity to IRC servers.
		*/
		class JUPITER_API Client : public Thinker
		{
		private:
			struct Data;
			Data *data_;
		protected:

			/**
			* @brief This is called when a connection has been successfully established.
			* The current implementation calls this after the MOTD.
			*/
			virtual void OnConnect();

			/**
			* @brief This is called when at the end of disconnect().
			* This is called upon any connection failure or error.
			*/
			virtual void OnDisconnect();

			/**
			* @brief This is called when at the end of reconnect().
			*
			* @param successConnect Used to determine if a connection was successfully established.
			* @see connect().
			*/
			virtual void OnReconnectAttempt(bool successConnect);

			/**
			* @brief This is called after a message has been normally processed.
			* This always happens last. All other hooks will be called before this.
			*
			* @param raw The raw message.
			*/
			virtual void OnRaw(const CStringType &raw);

			/**
			* @brief This is called after an IRC numeric has been processed.
			*
			* @param raw The raw message.
			*/
			virtual void OnNumeric(long int numeric, const CStringType &raw);

			/**
			* @brief This is called when an ERROR is received.
			* This indicates a connection termination, and thus, disconnect() is called immediately after this.
			*
			* @param message Message sent by the server.
			*/
			virtual void OnError(const CStringType &message);

			/**
			* @brief This is called when a chat message is received.
			*
			* @param channel String containing the destination of the message.
			* @param nick String containing the nickname of the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnChat(const CStringType &channel, const CStringType &nick, const CStringType &message);

			/**
			* @brief This is called when a notice is received.
			*
			* @param chan String containing the destination of the message.
			* @param nick String containing the nickname of the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnNotice(const CStringType &chan, const CStringType &sender, const CStringType &message);

			/**
			* @brief This is called when a server notice is received.
			*
			* @param chan String containing the destination of the message.
			* @param nick String containing the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnServerNotice(const CStringType &chan, const CStringType &sender, const CStringType &message);

			/**
			* @brief This is called when a CTCP message is received.
			*
			* @param channel String containing the destination of the message.
			* @param nick String containing the nickname of the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnCTCP(const CStringType &channel, const CStringType &nick, const CStringType &command, const CStringType &message);

			/**
			* @brief This is called when an action message is received.
			*
			* @param chan String containing the destination of the message.
			* @param nick String containing the nickname of the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnAction(const CStringType &chan, const CStringType &nick, const CStringType &message);

			/**
			* @brief This is called when an invite is received.
			*
			* @param chan String containing the destination of the message.
			* @param inviter String containing the nickname of the inviter.
			* @param invited String containing the nickname of the user invited.
			*/
			virtual void OnInvite(const CStringType &chan, const CStringType &inviter, const CStringType &invited);

			/**
			* @brief This is called when a chat message is received.
			*
			* @param chan String containing the destination of the message.
			* @param nick String containing the nickname of the sender.
			* @param message String containing the message sent.
			*/
			virtual void OnJoin(const CStringType &chan, const CStringType &nick);

			/**
			* @brief This is called when a user parts a channel.
			*
			* @param chan String containing the channel parted from.
			* @param nick String containing the nickname of the user.
			* @param reason String containing the reason for parting, or nullptr if none is specified.
			*/
			virtual void OnPart(const CStringType &chan, const CStringType &nick, const CStringType &reason);

			/**
			* @brief This is called when a user changes their nickname.
			*
			* @param oldnick String containing the old nickname of the user.
			* @param newnick String containing the new nickname of the user.
			*/
			virtual void OnNick(const CStringType &oldnick, const CStringType &newnick);

			/**
			* @brief This is called when a user is kicked from a channel.
			*
			* @param chan String containing the channel kicked from.
			* @param kicker String containing the nickname of the kicker.
			* @param kicked String containing the nickname of the user kicked.
			* @param reason String containing the reason for the kick, or nullptr if none is specified.
			*/
			virtual void OnKick(const CStringType &chan, const CStringType &kicker, const CStringType &kicked, const CStringType &reason);

			/**
			* @brief This is called when a user quits the server.
			*
			* @param nick String containing the nickname of the user.
			* @param message String containing the reason for quiting.
			*/
			virtual void OnQuit(const CStringType &nick, const CStringType &message);

			/**
			* @brief This is called when a channel mode is changed.
			*
			* @param chan String containing the affected channel.
			* @param nick String containing the nickname of the user.
			* @param modeString String containing the modes changed.
			*/
			virtual void OnMode(const Jupiter::CStringType &chan, const Jupiter::CStringType &nick, const Jupiter::CStringType &modeString);
		public:

			static INIFile *Config; /** IRC client config file. This should be instantiated before any Client instances. */

			/**
			* @brief Provides interfacing with channels and stores channel data.
			*/
			class JUPITER_API Channel
			{
			private:
				char *channel;
				char **users;
				char **userPrefixes;
				const char *serverPrefixSetPtr;
				int userCount;
				int userArraySize;
				int type;
			public:

				/**
				* @brief Returns the name of the channel.
				*
				* @return String containing the name of the channel.
				*/
				const char *getName();

				/**
				* @brief Returns a user at an index.
				*
				* @param index The index of the user.
				* @return A username if within range, nullptr otherwise.
				*/
				const char *getUser(short index);

				/**
				* @brief Adds a user to the channel
				*
				* @param nickname Nickname of the user.
				* @return Index of the new user.
				*/
				short addUser(const char *nickname);

				/**
				* @brief Adds a user to the channel
				*
				* @param nickname Nickname of the user.
				* @param prefix The user's prefix.
				* @return Index of the new user.
				*/
				short addUser(const char *nickname, const char prefix);

				/**
				* @brief Changes a user's nickname.
				*
				* @param oldnick String containing the old nickname.
				* @param newnick String containing the new nickname.
				* @return Index of the affected user.
				*/
				short changeUserName(const char *oldnick, const char *newnick);

				/**
				* @brief Changes a user's nickname.
				*
				* @param oldnick Index of a user.
				* @param newnick String containing the new nickname.
				* @return Index of the affected user.
				*/
				short changeUserName(short index, const char *newnick);

				/**
				* @brief Removes a user from the channel.
				*
				* @param nickname String containing the nickname of the user.
				*/
				void delUser(const char *nickname);

				/**
				* @brief Removes a user from the channel.
				*
				* @param index Index of a user.
				*/
				void delUser(short index);

				/**
				* @brief Adds a prefix to a user.
				*
				* @param index Index of a user.
				* @param prefix Prefix to add to the user.
				*/
				void addUserPrefix(short index, char prefix);

				/**
				* @brief Adds a prefix to a user.
				*
				* @param user String containing the nickname of the user.
				* @param prefix Prefix to add to the user.
				*/
				void addUserPrefix(const char *user, char prefix);

				/**
				* @brief Removes a prefix from a user.
				*
				* @param index Index of a user.
				* @param prefix Prefix to remove from the user.
				*/
				void delUserPrefix(short index, char prefix);

				/**
				* @brief Removes a prefix from a user.
				*
				* @param user String containing the nickname of a user.
				* @param prefix Prefix to remove from the user.
				*/
				void delUserPrefix(const char *user, char prefix);

				/**
				* @brief Returns the index of a user.
				*
				* @param user String containing the nickname of a user.
				* @return Index of a user.
				*/
				short getUserIndex(const char *user);

				/**
				* @brief Returns the index of a user.
				*
				* @param user String containing part of the nickname of a user.
				* @return Index of a user.
				*/
				short getUserIndexByPartName(const char *user);

				/**
				* @brief Returns the number of users in this channel.
				*
				* @return Number of users.
				*/
				int getUserCount();

				/**
				* @brief Returns the type of this channel.
				*
				* @return Type of this channel.
				*/
				int getType();

				/**
				* @brief Sets the type of this channel.
				*
				* @param iType Type to set this channel to.
				*/
				void setType(int iType);

				/**
				* @brief Returns a user's most significant prefix.
				*
				* @param index Index of a user.
				* @return User's most significant prefix.
				*/
				char getUserPrefix(int index);

				/**
				* @brief Returns a user's most significant prefix.
				*
				* @param user String containing the nickname of a user.
				* @return User's most significant prefix.
				*/
				char getUserPrefix(const char *user);

				bool isAddingNames;

				/**
				* @brief Constructor for Channel
				*
				* @param channelName String containing the name of a channel.
				* @param iFace Server in which this channel is located.
				*/
				Channel(const char *channelName, Client *iFace);

				/**
				* @brief Destructor for Channel
				*/
				~Channel();
			}; // Jupiter::IRC::Client::Channel class

			/**
			* @brief Returns the config section this refers to.
			*
			* @return String containing a config section.
			*/
			const char *getConfigSection();

			/**
			* @brief Returns the name of the file this logs to.
			*
			* @return String containing a log file's name.
			*/
			const char *getLogFile();

			/**
			* @brief Returns the nickname prefixes supported by the connected server.
			*
			* @return String containing nickname prefixes.
			*/
			const char *getPrefixes();

			/**
			* @brief Returns the client's current nickname.
			*
			* @return String containing a nickame.
			*/
			const char *getNickname();

			/**
			* @brief Returns the client's real name.
			*
			* @return String containing a name.
			*/
			const char *getRealname();

			/**
			* @brief Returns the server's name
			*
			* @return String containing the server's name.
			*/
			const char *getServerName();

			/**
			* @brief Returns the server's hostname.
			* Note: This is the same as the config's value.
			*
			* @return String containing a hostname.
			*/
			const char *getServerHostname();

			/**
			* @brief Returns the server's port.
			* Note: This is the same as the config's value.
			*
			* @return Port used to connect.
			*/
			unsigned short getServerPort();

			/**
			* @brief Returns the time delay between reconnect attempts.
			*
			* @return Time delay between reconnect attemps.
			*/
			time_t getReconnectDelay();

			/**
			* @brief Returns the time scheduled to make a reconenct attempt.
			* This is the sum of the time of disconnection and the reconnection time delay.
			*
			* @return Time scheduled.
			*/
			time_t getReconnectTime();

			/**
			* @brief Returns the number of consecutive reconnect attempts executed so far.
			*
			* @return Number of reconnects attempted.
			*/
			int getReconnectAttempts();

			/**
			* @brief Returns the maximum number of consecutive reconnect attempts to make before failing.
			*
			* @return Maximum number of reconnect attempts.
			*/
			int getMaxReconnectAttempts();

			/**
			* @brief Returns the default channel type.
			*
			* @return Default channel type.
			*/
			short getDefaultChanType();

			/**
			* @brief Returns if this prints to stdout.
			*
			* @return True if it prints, false otherwise.
			*/
			FILE *getPrintOutput();

			/**
			* @brief Sets where to print output.
			*
			* @param outf FILE object to output to.
			* Note: outf must be either a valid and open FILE pointer, or nullptr to disable output.
			*/
			void setPrintOutput(FILE *outf);

			/**
			* @brief Returns the number of channels.
			*
			* @return Number of channels.
			*/
			int getChannelCount();

			/**
			* @brief Returns the channel array size.
			*
			* @return Channel array size.
			*/
			int getChannelsSize();

			/**
			* @brief Returns a channel.
			*
			* @param index Index of the chanel to return.
			* @return The channel at index.
			*/
			Channel *getChannel(short index);

			/**
			* @brief Returns a channel.
			*
			* @param chanName String containing the name of a channel.
			* @return The channel at index.
			*/
			Channel *getChannel(const char *chanName);

			/**
			* @brief Returns a channel's name at index.
			*
			* @param index Index of the channel.
			* @return Channel's name.
			*/
			const char *getChannelName(short index);

			/**
			* @brief Returns the index of the channel with a given name.
			*
			* @param chanName String containing the name of a channel.
			* @return Index of a channel.
			*/
			short getChannelIndex(const char *chanName);

			/**
			* @brief Sends a join request.
			*
			* @param channel Channel to join.
			*/
			void joinChannel(const char *channel);

			/**
			* @brief Sends a join request with a password.
			*
			* @param channel Channel to join.
			* @param password Password to use.
			*/
			void joinChannel(const char *channel, const char *password);

			/**
			* @brief Parts a channel.
			*
			* @param channel Channel to part.
			*/
			void partChannel(const char *channel);

			/**
			* @brief Parts a channel.
			*
			* @param channel Channel to part.
			* @param message Reason for parting.
			*/
			void partChannel(const char *channel, const char *message);

			/**
			* @brief Gets the access level of a user.
			*
			* @param chan String containing a channel name to get the access level from.
			* @param nick String containing the nickname of the user.
			* @return Access level of the user.
			*/
			int getAccessLevel(const char *chan, const char *nick);

			/**
			* @brief Sends a message.
			*
			* @param dest String containing the destination of the message (nickname or channel).
			* @param message String containing the message to send.
			*/
			void sendMessage(const char *dest, const char *message);

			/**
			* @brief Sends a notice.
			*
			* @param dest String containing the destination of the message (nickname or channel).
			* @param message String containing the message to send.
			*/
			void sendNotice(const char *dest, const char *message);

			/**
			* @brief Sends a message to all channels of a given type.
			*
			* @param type Type of channel to messasge.
			* @param message String containing the message to send.
			* @return Number of messages sent.
			*/
			int messageChannels(short type, const char *message);

			/**
			* @brief Sends a message to all channels with a type of at least 0.
			*
			* @param message String containing the message to send.
			* @return Number of messages sent.
			*/
			int messageChannels(const char *message);

			/**
			* @brief Returns if the client will automatically reconnect upon failure.
			*
			* @return True if the server will automatically reconnect, false otherwise.
			*/
			bool isAutoReconnect();

			/**
			* @brief Sets how many times the client should attempt to automatically reconnect.
			*
			* @param autoReconnect Maximum number of reconnect attempts. A negative value indicates infinity.
			*/
			void setAutoReconnect(int autoReconnect);

			/**
			* @brief Sends data to the server.
			* Endlines are automatically added.
			*
			* @param rawMessage String containing the data to send.
			*/
			void send(const char *rawMessage);

			/**
			* @brief Method that is called by think() to handle all of the IRC protocol.
			* Note: This command should not be used unless you wish to bypass auto-reconnect abilities. Use think() instead.
			*
			* @return 0 upon success, an error code otherwise.
			*/
			int primaryHandler();

			/**
			* @brief Returns a key's value.
			* This reads from the client's config section first, then default if it doesn't exist.
			*
			* @param key String containing the key name.
			* @return String containing the key value if it exists, nullptr otherwise.
			*/
			char *readConfigValue(const char *key);

			/**
			* @brief Returns a key's value as a boolean.
			* This reads from the client's config section first, then default if it doesn't exist.
			*
			* @param key String containing the key name.
			* @return Boolean value of the key value if it exists, false otherwise.
			*/
			bool readConfigBool(const char *key);

			/**
			* @brief Returns a key's value as an integer.
			* This reads from the client's config section first, then default if it doesn't exist.
			*
			* @param key String containing the key name.
			* @return Integer value of the key value if it exists, 0 otherwise.
			*/
			int readConfigInt(const char *key);

			/**
			* @brief Returns a key's value as a long integer.
			* This reads from the client's config section first, then default if it doesn't exist.
			*
			* @param key String containing the key name.
			* @return Long integer value of the key value if it exists, 0 otherwise.
			*/
			long readConfigLong(const char *key);

			/**
			* @brief Returns a key's value as a double.
			* This reads from the client's config section first, then default if it doesn't exist.
			*
			* @param key String containing the key name.
			* @return Double value of the key value if it exists, 0 otherwise.
			*/
			double readConfigDouble(const char *key);

			/**
			* @brief Writes to a server's log file.
			*
			* @param message String containing the text to write to the file.
			*/
			void writeToLogs(const char *message);

			/**
			* @brief Connects the client to its server.
			* Note: This should not be called unless it is not already connected to a server.
			* Also, a successful socket connection does not mean the server accepted the connection.
			*
			* @return True upon successful socket connection, false otherwise.
			*/
			bool connect();

			/**
			* @brief Completely disconnects the client from a server.
			* Note: This does not send any type of QUIT message.
			*/
			void disconnect(bool stayDead = false);

			/**
			* @brief Completely disconnects the client from a server.
			*
			* @param message String containing the QUIT message to send prior to disconnecting.
			*/
			void disconnect(const char *message, bool stayDead = false);

			/**
			* @brief Calls disconnect() if the client has not already, then calls connect().
			* Note: This will increment the current reconnect attempts by 1.
			*/
			void reconnect();

			/**
			* @brief Method to be looped to sustain the IRC connection.
			* Note: This client should be deleted if this returns anything other than 0.
			*
			* @return 0 if the client should still exist, error code otherwise.
			*/
			virtual int think();

			/**
			* @brief Constructor for a client.
			*
			* @param configSection String containing the config section for the client to read from, before defaulting to "Default".
			*/
			Client(const char *configSection);

			/**
			* @brief Destructor for a client.
			*/
			virtual ~Client();

		}; // Jupiter::IRC::Client class

	} // Jupiter::IRC namespace

} // Jupiter namespace

#endif // _IRC_CLIENT_H_HEADER