/**
 * Copyright (C) 2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#include "INIFile.h"
#include "ServerManager.h"
#include "IRC_Bot.h"
#include "RenX_Server.h"
#include "RenX_PlayerInfo.h"
#include "RenX_GameCommand.h"
#include "RenX_Functions.h"

int RenX::Server::send(const char *command, size_t length)
{
	char *t = new char[length + 3];
	*t = 'c';
	Jupiter::strcpy<char>(t + 1, command, length);
	t[length + 1] = '\n';
	t[length + 2] = 0;
	int r;
	RenX::sanitizeString(t);
	r = RenX::Server::sock.send(t, length + 3);
	delete[] t;
	return r;
}

int RenX::Server::send(const char *command)
{
	return RenX::Server::send(command, strlen(command));
}

int RenX::Server::sendData(const char *data, size_t size)
{
	return RenX::Server::sock.send(data, size);
}

int RenX::Server::sendData(const char *data)
{
	return RenX::Server::sock.send(data, strlen(data));
}

RenX::PlayerInfo *RenX::Server::getPlayer(int id)
{
	if (RenX::Server::players.size() == 0) return nullptr;
	for (Jupiter::DLList<RenX::PlayerInfo>::Node *node = RenX::Server::players.getNode(0); node != nullptr; node = node->next) if (node->data->id == id) return node->data;
	return nullptr;
}

RenX::PlayerInfo *RenX::Server::getPlayerByName(const Jupiter::CStringType &name)
{
	if (RenX::Server::players.size() == 0) return nullptr;
	for (Jupiter::DLList<RenX::PlayerInfo>::Node *node = RenX::Server::players.getNode(0); node != nullptr; node = node->next) if (node->data->name == name) return node->data;
	return nullptr;
}

RenX::PlayerInfo *RenX::Server::getPlayerByPartName(const Jupiter::CStringType &partName)
{
	if (RenX::Server::players.size() == 0) return nullptr;
	RenX::PlayerInfo *r = RenX::Server::getPlayerByName(partName);
	if (r != nullptr) return r;
	return RenX::Server::getPlayerByPartNameFast(partName);
}

RenX::PlayerInfo *RenX::Server::getPlayerByPartNameFast(const Jupiter::CStringType &partName)
{
	if (RenX::Server::players.size() == 0) return nullptr;
	const char *cStr = partName.c_str();
	for (Jupiter::DLList<RenX::PlayerInfo>::Node *node = RenX::Server::players.getNode(0); node != nullptr; node = node->next) if (stristr(node->data->name.c_str(), cStr)) return node->data;
	return nullptr;
}

bool RenX::Server::removePlayer(int id)
{
	if (RenX::Server::players.size() == 0) return false;
	for (Jupiter::DLList<RenX::PlayerInfo>::Node *node = RenX::Server::players.getNode(0); node != nullptr; node = node->next)
	{
		if (node->data->id == id)
		{
			RenX::PlayerInfo *p = RenX::Server::players.remove(node);
			delete p;
			return true;
		}
	}
	return false;
}

bool RenX::Server::removePlayer(RenX::PlayerInfo *player)
{
	return RenX::Server::removePlayer(player->id);
}

const Jupiter::CStringType &RenX::Server::getPrefix() const
{
	return RenX::Server::IRCPrefix;
}

void RenX::Server::setPrefix(const char *prefix)
{
	RenX::Server::IRCPrefix = prefix;
}

const Jupiter::CStringType &RenX::Server::getCommandPrefix() const
{
	return RenX::Server::CommandPrefix;
}

void RenX::Server::setCommandPrefix(const char *prefix)
{
	RenX::Server::CommandPrefix = prefix;
}

RenX::GameCommand *RenX::Server::getCommand(unsigned int index)
{
	return RenX::Server::commands.get(index);
}

RenX::GameCommand *RenX::Server::getCommand(const char *trigger)
{
	for (int i = RenX::Server::commands.size() - 1; i >= 0; i--)
	{
		RenX::GameCommand *cmd = RenX::Server::commands.get(i);
		if (cmd->matches(trigger)) return cmd;
	}
	return nullptr;
}

unsigned int RenX::Server::getCommandCount()
{
	return RenX::Server::commands.size();
}

unsigned int RenX::Server::triggerCommand(const char *trigger, RenX::PlayerInfo *player, const Jupiter::CStringS &parameters)
{
	unsigned int r = 0;
	RenX::GameCommand *cmd;
	for (int i = RenX::Server::commands.size() - 1; i >= 0; i--)
	{
		cmd = RenX::Server::commands.get(i);
		if (cmd->matches(trigger))
		{
			cmd->trigger(this, player, parameters);
			r++;
		}
	}
	return r;
}

void RenX::Server::addCommand(RenX::GameCommand *command)
{
	RenX::Server::commands.add(command);
}

bool RenX::Server::removeCommand(RenX::GameCommand *command)
{
	for (int i = RenX::Server::commands.size() - 1; i >= 0; i--)
	{
		if (RenX::Server::commands.get(i) == command)
		{
			delete RenX::Server::commands.remove(i);
			return true;
		}
	}
	return false;
}

bool RenX::Server::removeCommand(const char *trigger)
{
	for (int i = RenX::Server::commands.size() - 1; i >= 0; i--)
	{
		if (RenX::Server::commands.get(i)->matches(trigger))
		{
			delete RenX::Server::commands.remove(i);
			return true;
		}
	}
	return false;
}

void RenX::Server::sendPubChan(const char *fmt, ...) const
{
	va_list args;
	va_start(args, fmt);
	Jupiter::CStringL msg;
	const Jupiter::CStringS &serverPrefix = RenX::Server::IRCPrefix;
	if (serverPrefix != "")
	{
		msg += serverPrefix;
		msg += ' ';
		msg.avformat(fmt, args);
	}
	else msg.vformat(fmt, args);
	va_end(args);
	for (int b = serverManager->size() - 1; b >= 0; b--) serverManager->getServer(b)->messageChannels(RenX::Server::logChanType, msg.c_str());
}

void RenX::Server::sendAdmChan(const char *fmt, ...) const
{
	va_list args;
	va_start(args, fmt);
	Jupiter::CStringL msg;
	const Jupiter::CStringS &serverPrefix = RenX::Server::getPrefix();
	if (serverPrefix != "")
	{
		msg += serverPrefix;
		msg += ' ';
		msg.avformat(fmt, args);
	}
	else msg.vformat(fmt, args);
	va_end(args);
	for (int b = serverManager->size() - 1; b >= 0; b--) serverManager->getServer(b)->messageChannels(RenX::Server::adminLogChanType, msg.c_str());
}

void RenX::Server::sendLogChan(const char *fmt, ...) const
{
	va_list args;
	va_start(args, fmt);
	Jupiter::CStringL msg;
	const Jupiter::CStringS &serverPrefix = RenX::Server::getPrefix();
	if (serverPrefix != "")
	{
		msg += serverPrefix;
		msg += ' ';
		msg.avformat(fmt, args);
	}
	else msg.vformat(fmt, args);
	va_end(args);
	for (int b = serverManager->size() - 1; b >= 0; b--)
	{
		IRC_Bot *ircServer = serverManager->getServer(b);
		ircServer->messageChannels(RenX::Server::logChanType, msg.c_str());
		ircServer->messageChannels(RenX::Server::adminLogChanType, msg.c_str());
	}
}

bool RenX::Server::connect()
{
	if (RenX::Server::sock.connectToHost(RenX::Server::hostname, RenX::Server::port, RenX::Server::clientHostname))
	{
		RenX::Server::sock.setBlocking(false);
		size_t tBuffSize = strlen(RenX::Server::pass) + 32;
		char *tBuff = new char[tBuffSize];
		RenX::Server::sock.send(tBuff, sprintf(tBuff, "a%s\n", RenX::Server::pass));
		delete[] tBuff;
		RenX::Server::sock.send("s\n", 2);
		RenX::Server::sock.send("clogclientlist\n", 15);
		return true;
	}
	return false;
}

RenX::Server::Server(Jupiter::CString_Type<char> &configurationSection)
{
	RenX::Server::configSection = configurationSection;
	const char *tPtr;

	tPtr = Jupiter::IRC::Client::Config->get(RenX::Server::configSection.c_str(), "Hostname");
	if (tPtr == nullptr) tPtr = "localhost";
	RenX::Server::hostname = makestr(tPtr);

	RenX::Server::port = Jupiter::IRC::Client::Config->getShort(RenX::Server::configSection.c_str(), "Port");
	if (RenX::Server::port == 0) RenX::Server::port = 7777;

	tPtr = Jupiter::IRC::Client::Config->get(RenX::Server::configSection.c_str(), "Password");
	if (tPtr == nullptr) tPtr = "renx";
	RenX::Server::pass = makestr(tPtr);

	RenX::Server::logChanType = Jupiter::IRC::Client::Config->getShort(RenX::Server::configSection.c_str(), "ChanType");
	RenX::Server::adminLogChanType = Jupiter::IRC::Client::Config->getShort(RenX::Server::configSection.c_str(), "AdminChanType");

	RenX::Server::clientHostname = Jupiter::IRC::Client::Config->get(RenX::Server::configSection.c_str(), "ClientAddress");
	
	tPtr = Jupiter::IRC::Client::Config->get(RenX::Server::configSection.c_str(), "CommandPrefix");
	if (tPtr != nullptr) RenX::Server::setCommandPrefix(tPtr);

	tPtr = Jupiter::IRC::Client::Config->get(RenX::Server::configSection.c_str(), "IRCPrefix");
	if (tPtr != nullptr) RenX::Server::setPrefix(tPtr);

	for (int i = RenX::GameMasterCommandList->size() - 1; i >= 0; i--) RenX::Server::addCommand(RenX::GameMasterCommandList->get(i)->copy());
}

RenX::Server::~Server()
{
	sock.closeSocket();
	if (RenX::Server::hostname != nullptr) delete[] RenX::Server::hostname;
	if (RenX::Server::pass != nullptr) delete[] RenX::Server::pass;
	while (RenX::Server::players.size() > 0) delete RenX::Server::players.remove(unsigned int(0));
	for (int i = RenX::Server::commands.size() - 1; i >= 0; i--) delete RenX::Server::commands.remove(i);
}
