/**
 * Copyright (C) 2014 Justin James.
 *
 * This license must be preserved.
 * Any applications, libraries, or code which make any use of any
 * component of this program must not be commercial, unless explicit
 * permission is granted from the original author. The use of this
 * program for non-profit purposes is permitted.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * In the event that this license restricts you from making desired use of this program, contact the original author.
 * Written by Justin James <justin.aj@hotmail.com>
 */

#if !defined RENX_SERVER_H_HEADER
#define RENX_SERVER_H_HEADER

/**
 * @file RenX_Server.h
 * @brief Defines the Server class.
 */

#include "RenX.h"
#include "TCPSocket.h"
#include "DLList.h"
#include "ArrayList.h"
#include "CString.h"

namespace RenX
{
	/** Forward declarations */
	struct PlayerInfo;
	class GameCommand;

	/** Template classes */
	template class RENX_API Jupiter::ArrayList<RenX::GameCommand>;
	template class RENX_API Jupiter::DLList<RenX::PlayerInfo>;
	template class RENX_API Jupiter::CString_Strict<char>;

	/**
	* @brief Represents a connection to an individiaul Renegade-X server.
	* There are often more than one of these, such as when communities run multiple servers.
	*/
	class RENX_API Server
	{
	public:
		Jupiter::DLList<RenX::PlayerInfo> players;
		Jupiter::TCPSocket sock;
		unsigned short port = 0;
		const char *hostname = nullptr;
		const char *pass = nullptr;
		const char *clientHostname = nullptr;
		short logChanType;
		short adminLogChanType;
		// Change above to private.

		/**
		* @brief Sends a command to the server.
		*
		* @param commmand Command to send.
		* @param size Length of the string, not including the null terminator.
		* @return The number of bytes sent on success, less than or equal to zero otherwise.
		*/
		int send(const char *command, size_t size);

		/**
		* @brief Sends a command to the server.
		*
		* @param commmand Command to send.
		* @return The number of bytes sent on success, less than or equal to zero otherwise.
		*/
		int send(const char *command);

		/**
		* @brief Sends data to the server.
		*
		* @param data Data to send.
		* @param size Length of the data in bytes.
		* @return The number of bytes sent on success, less than or equal to zero otherwise.
		*/
		int sendData(const char *data, size_t size);

		/**
		* @brief Sends data to the server.
		*
		* @param data Data in the form of a C-style string.
		* @return The number of bytes sent on success, less than or equal to zero otherwise.
		*/
		int sendData(const char *data);

		/**
		* @brief Fetches a player's data based on their ID number.
		*
		* @param id ID of the player
		* @return A pointer to a player's corresponding PlayerInfo on success, nullptr otherwise.
		*/
		RenX::PlayerInfo *getPlayer(int id);

		/**
		* @brief Fetches a player's data, based on their name.
		*
		* @param name Name of the player.
		* @return A player's data on success, nullptr otherwise.
		*/
		RenX::PlayerInfo *getPlayerByName(const Jupiter::CStringType &name);

		/**
		* @brief Fetches a player's data, based on part of their name.
		*
		* @param name Name of the player.
		* @return A player's data on success, nullptr otherwise.
		*/
		RenX::PlayerInfo *getPlayerByPartName(const Jupiter::CStringType &partName);

		/**
		* @brief Fetches a player's data, based on part of their name.
		* Note: This does not check that the exact name exists first.
		*
		* @param name Name of the player.
		* @return A player's data on success, nullptr otherwise.
		*/
		RenX::PlayerInfo *getPlayerByPartNameFast(const Jupiter::CStringType &partName);

		/**
		* @brief Removes a player's data based on their ID number.
		*
		* @param id ID of the player
		* @return True if a player was removed, false otherwise.
		*/
		bool removePlayer(int id);

		/**
		* @brief Removes a player's data based on a player's data.
		* Note: Current implementation just returns removePlayer(player->id);
		*
		* @param player Pointer to a player's data
		* @return True if a player was removed, false otherwise.
		*/
		bool removePlayer(RenX::PlayerInfo *player);

		/**
		* @brief Fetches a server's IRC logging prefix.
		*
		* @return The server's logging prefix.
		*/
		const Jupiter::CStringType &getPrefix() const;

		/**
		* @brief Sets a server's IRC logging prefix.
		*
		* @param prefix Prefix to use.
		*/
		void setPrefix(const char *prefix);

		/**
		* @brief Fetches a server's game command prefix.
		*
		* @return The server's game command prefix.
		*/
		const Jupiter::CStringType &getCommandPrefix() const;

		/**
		* @brief Sets a server's game command prefix.
		*
		* @param prefix Prefix to use.
		*/
		void setCommandPrefix(const char *prefix);

		/**
		* @brief Fetches a command from the list.
		*
		* @param index Index of the command to return.
		* @return Command located at the specified index.
		*/
		GameCommand *getCommand(unsigned int index);

		/**
		* @brief Fetches a command from the list.
		*
		* @param trigger Trigger of the command to get
		* @return Command with the specified trigger.
		*/
		GameCommand *getCommand(const char *trigger);

		/**
		* @brief Returns the number of commands in the command list.
		*
		* @return Number of commands in the command list.
		*/
		unsigned int getCommandCount();

		/**
		* @brief Triggers commands matching a specified trigger.
		*
		* @param trigger Trigger of the command to fire.
		* @return Number of commands triggered.
		*/
		unsigned int triggerCommand(const char *trigger, RenX::PlayerInfo *player, const Jupiter::CStringS &parameters);

		/**
		* @brief Adds a command to the server's game command list.
		*
		* @param command Command to add.
		*/
		void addCommand(RenX::GameCommand *command);

		/**
		* @brief Adds a command to the server's game command list.
		*
		* @param command Command to remove.
		* @return True if a command was removed, false otherwise.
		*/
		bool removeCommand(RenX::GameCommand *command);

		/**
		* @brief Adds a command to the server's game command list.
		*
		* @param trigger Trigger of the command to remove.
		* @return True if a command was removed, false otherwise.
		*/
		bool removeCommand(const char *trigger);

		/**
		* @brief Formats and sends a message to a server's corresponding public channels.
		*
		* @param server Server corresponding to channel types.
		* @param fmt String containing the format specifiers indicating what message to send.
		*/
		void sendPubChan(const char *fmt, ...) const;

		/**
		* @brief Formats and sends a message to a server's corresponding adminstrative channels.
		*
		* @param server Server corresponding to channel types.
		* @param fmt String containing the format specifiers indicating what message to send.
		*/
		void sendAdmChan(const char *fmt, ...) const;

		/**
		* @brief Formats and sends a message to a server's corresponding channels.
		*
		* @param server Server corresponding to channel types.
		* @param fmt String containing the format specifiers indicating what message to send.
		*/
		void sendLogChan(const char *fmt, ...) const;

		/**
		* @brief Connects to the server's RCON interface.
		*
		* @return True if the socket successfully connected, false otherwise.
		*/
		bool connect();

		/**
		* @brief Creates the Server object, and loads settings from the specified configuration section.
		*
		* @param configurationSection Configuration section to load from.
		*/
		Server(Jupiter::CString_Type<char> &configurationSection);

		/**
		* @brief Destructor for Server class.
		*/
		~Server();

	/** Private members */
	private:
		Jupiter::CStringS configSection;
		Jupiter::CStringS rules;
		Jupiter::CStringS IRCPrefix;
		Jupiter::CStringS CommandPrefix;
		Jupiter::ArrayList<RenX::GameCommand> commands;
	};

}

#endif // RENX_SERVER_H_HEADER